/*
 * Copyright © 2012 Michael Heimpold <mhei@heimpold.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>

#include <config.h>
#include <ugpio.h>
#include <ugpio-internal.h>

static int ugpio_fd_open(unsigned int gpio, int ro)
{
    char pathname[255];
    snprintf(pathname, sizeof(pathname), GPIO_VALUE, gpio);

    return open(pathname, (ro ? O_RDONLY : O_RDWR) | O_NONBLOCK);
}

static int ugpio_fd_close(int fd)
{
    return close(fd);
}

gpio_t *ugpio_request_one(unsigned int gpio, unsigned int flags, const char *label)
{
    gpio_t *ctx;
    int is_requested;

    if ((ctx = malloc(sizeof(*ctx))) == NULL)
        return NULL;

    ctx->gpio = gpio;
    ctx->flags = flags;
    ctx->label = label;
    ctx->fd = -1;

    if ((is_requested = gpio_is_requested(ctx->gpio)) < 0)
        goto error_free;

    if (!is_requested) {
        if (gpio_request_one(ctx->gpio, ctx->flags, ctx->label) < 0)
            goto error_free;

        ctx->flags |= GPIOF_REQUESTED;
    }

    return ctx;

error_free:
    free(ctx);
    return NULL;
}

void ugpio_free(gpio_t *ctx)
{
    if (ctx == NULL)
        return;

    if (ctx->flags & GPIOF_REQUESTED)
        gpio_free(ctx->gpio);

    free(ctx);
}

int ugpio_open(gpio_t *ctx)
{
    ctx->fd = ugpio_fd_open(ctx->gpio, ctx->flags & GPIOF_DIR_IN);
    return ctx->fd;
}

void ugpio_close(gpio_t *ctx)
{
    if (ctx == NULL)
        return;

    if (ctx->fd == -1)
        return;

    ugpio_fd_close(ctx->fd);
    ctx->fd = -1;
}

int ugpio_fd(gpio_t *ctx)
{
    return ctx->fd;
}

int ugpio_get_value(gpio_t *ctx)
{
    char buffer;
    ssize_t c;
    off_t o;

    if ((o = lseek(ctx->fd, 0, SEEK_SET)) == -1)
        return -1;

    if ((c = read(ctx->fd, &buffer, sizeof(buffer))) == -1)
        return -1;

    return buffer - '0';
}

int ugpio_set_value(gpio_t *ctx, int value)
{
    ssize_t c;

    c = write(ctx->fd, value ? "1" : "0", 2);

    return (c != 2) ? -1 : 0;
}
